@@ -4,6 +4,21 @@
  ****************************************************************************/
 
 #include "jwm.h"
+#include "client.h"
+#include "main.h"
+#include "icon.h"
+#include "hint.h"
+#include "group.h"
+#include "tray.h"
+#include "confirm.h"
+#include "key.h"
+#include "cursor.h"
+#include "taskbar.h"
+#include "screen.h"
+#include "pager.h"
+#include "color.h"
+#include "error.h"
+#include "place.h"
 
 static const int STACK_BLOCK_SIZE = 8;
 
@@ -16,23 +31,13 @@ static Window *stack;
 static int stackSize;
 static int clientCount;
 
-static unsigned int cascadeOffset;
-static unsigned int cascadeStart;
-static unsigned int cascadeStop;
-
 static void LoadFocus();
 static void ReparentClient(ClientNode *np, int notOwner);
  
-static void SendClientMessage(ClientNode *np, AtomType type,
-	AtomType message);
-static void GetBorderOffsets(ClientNode *np, int *north, int *west);
-static void PlaceWindow(ClientNode *np, int alreadyMapped);
-static void Gravitate(ClientNode *np, int negate);
-static void UpdateState(ClientNode *np);
 static void MinimizeTransients(ClientNode *np);
 static void CheckShape(ClientNode *np);
 
-static void RestoreSingleClient(ClientNode *np);
+static void RestoreTransients(ClientNode *np);
 
 static void KillClientHandler(ClientNode *np);
 
@@ -48,7 +53,7 @@ void StartupClients() {
 	XWindowAttributes attr;
 	Window rootReturn, parentReturn, *childrenReturn;
 	unsigned int childrenCount;
-	int x;
+	unsigned int x;
 
 	JXSync(display, False);
 	JXGrabServer(display);
@@ -59,16 +64,11 @@ void StartupClients() {
 	activeClient = NULL;
 	currentDesktop = 0;
 
-	cascadeStart = titleHeight + borderWidth;
-	cascadeStop = rootHeight / 3;
-	cascadeOffset = cascadeStart;
-
 	for(x = 0; x < LAYER_COUNT; x++) {
 		nodes[x] = NULL;
 		nodeTail[x] = NULL;
 	}
 
-
 	JXQueryTree(display, rootWindow, &rootReturn, &parentReturn,
 		&childrenReturn, &childrenCount);
 
@@ -90,6 +90,9 @@ void StartupClients() {
 
 	ExposeCurrentDesktop();
 
+	UpdateTaskBar();
+	UpdatePager();
+
 }
 
 /****************************************************************************
@@ -136,8 +139,8 @@ void LoadFocus() {
  * Add a window to management.
  ****************************************************************************/
 ClientNode *AddClientWindow(Window w, int alreadyMapped, int notOwner) {
+
 	XWindowAttributes attr;
-	XWMHints *wmhints;
 	ClientNode *np;
 
 	Assert(w != None);
@@ -158,7 +161,7 @@ ClientNode *AddClientWindow(Window w, int alreadyMapped, int notOwner) {
 
 	np->window = w;
 	np->owner = None;
-	np->desktop = currentDesktop;
+	np->state.desktop = currentDesktop;
 	np->controller = NULL;
 	np->name = NULL;
 	np->colormaps = NULL;
@@ -169,67 +172,49 @@ ClientNode *AddClientWindow(Window w, int alreadyMapped, int notOwner) {
 	np->height = attr.height;
 	np->cmap = attr.colormap;
 	np->colormaps = NULL;
-	np->statusFlags = STAT_NONE;
-	np->layer = LAYER_NORMAL;
+	np->state.status = STAT_NONE;
+	np->state.layer = LAYER_NORMAL;
 
 	if(notOwner) {
-		np->borderFlags = BORDER_OUTLINE | BORDER_TITLE
-			| BORDER_MIN | BORDER_MAX | BORDER_CLOSE
-			| BORDER_MOVE | BORDER_RESIZE;
+		np->state.border = BORDER_DEFAULT;
 	} else {
-		np->borderFlags = BORDER_OUTLINE | BORDER_TITLE | BORDER_MOVE;
-		np->statusFlags |= STAT_WMDIALOG | STAT_STICKY;
+		np->state.border = BORDER_OUTLINE | BORDER_TITLE | BORDER_MOVE;
+		np->state.status |= STAT_WMDIALOG | STAT_STICKY;
 	}
 	np->borderAction = BA_NONE;
 
-	ReadMotifHints(np);
 	ReadClientProtocols(np);
 
 	/* We now know the layer, so insert */
 	np->prev = NULL;
-	np->next = nodes[np->layer];
+	np->next = nodes[np->state.layer];
 	if(np->next) {
 		np->next->prev = np;
 	} else {
-		nodeTail[np->layer] = np;
+		nodeTail[np->state.layer] = np;
 	}
-	nodes[np->layer] = np;
+	nodes[np->state.layer] = np;
 
-	wmhints = XGetWMHints(display, np->window);
-	if(wmhints) {
-		switch(wmhints->flags & StateHint) {
-		case IconicState:
-			np->statusFlags |= STAT_MINIMIZED;
-			break;
-		default:
-			if(!(np->statusFlags & STAT_MINIMIZED)) {
-				np->statusFlags |= STAT_MAPPED;
-			}
-			break;
-		}
-		JXFree(wmhints);
-	} else {
-		np->statusFlags |= STAT_MAPPED;
-	}
+	LoadIcon(np);
+
+	ApplyGroups(np);
 
 	SetDefaultCursor(np->window);
 	ReparentClient(np, notOwner);
-	PlaceWindow(np, alreadyMapped);
+	PlaceClient(np, alreadyMapped);
 
 	/* If one of these fails we are SOL, so who cares. */
 	XSaveContext(display, np->window, clientContext, (void*)np);
 	XSaveContext(display, np->parent, frameContext, (void*)np);
 
-	if(np->statusFlags & STAT_MAPPED) {
+	if(np->state.status & STAT_MAPPED) {
 		JXMapWindow(display, np->window);
 		JXMapWindow(display, np->parent);
 	}
 
 	DrawBorder(np);
 
-	LoadIcon(np);
-
-	AddClientToTray(np);
+	AddClientToTaskBar(np);
 
 	if(!alreadyMapped) {
 		RaiseClient(np);
@@ -237,165 +222,41 @@ ClientNode *AddClientWindow(Window w, int alreadyMapped, int notOwner) {
 
 	++clientCount;
 
-	ApplyGroups(np);
-
-	if(np->statusFlags & STAT_STICKY) {
-		SetCardinalAtom(np->window, ATOM_NET_WM_DESKTOP, 0xFFFFFFFF);
+	if(np->state.status & STAT_STICKY) {
+		SetCardinalAtom(np->window, ATOM_NET_WM_DESKTOP, ~0UL);
 	} else {
-		SetCardinalAtom(np->window, ATOM_NET_WM_DESKTOP, np->desktop);
+		SetCardinalAtom(np->window, ATOM_NET_WM_DESKTOP, np->state.desktop);
 	}
 
-	if(np->statusFlags & STAT_SHADED) {
+	if(np->state.status & STAT_SHADED) {
 		ShadeClient(np);
 	}
 
-	/* Make sure we're still in sync */
-	UpdateState(np);
-	SendConfigureEvent(np);
-	WriteWinState(np);
+	if(np->state.status & STAT_MINIMIZED) {
+		np->state.status &= ~STAT_MINIMIZED;
+		MinimizeClient(np);
+	}
 
-	if(np->desktop != currentDesktop && !(np->statusFlags & STAT_STICKY)) {
+	if(np->state.status & STAT_MAXIMIZED) {
+		np->state.status &= ~STAT_MAXIMIZED;
+		MaximizeClient(np);
+	}
+
+	/* Make sure we're still in sync */
+	WriteState(np);
+	SendConfigureEvent(np);
+
+	if(np->state.desktop != currentDesktop
+		&& !(np->state.status & STAT_STICKY)) {
 		HideClient(np);
 	}
 
+	ReadClientStrut(np);
+
 	return np;
 
 }
 
-/****************************************************************************
- * Place a window on the screen.
- ****************************************************************************/
-void PlaceWindow(ClientNode *np, int alreadyMapped) {
-	int north, west;
-
-	Assert(np);
-
-	GetBorderOffsets(np, &north, &west);
-
-	if(!(np->sizeFlags & PPosition) && !alreadyMapped) {
-
-		if(np->width + cascadeOffset + west > rootWidth) {
-			if(np->width + west * 2 < rootWidth) {
-				np->x = west + rootWidth / 2 - np->width / 2;
-			} else {
-				np->x = west;
-			}
-		} else {
-			np->x = west + cascadeOffset;
-		}
-
-		if(np->height + cascadeOffset + north > rootHeight - trayHeight) {
-			if(np->height + north + west < rootHeight) {
-				np->y = north + rootWidth / 2 - np->width / 2;
-			} else {
-				np->y = north;
-			}
-		} else {
-			np->y = north + cascadeOffset;
-		}
-
-		if(cascadeOffset >= cascadeStop) {
-			cascadeOffset = cascadeStart;
-		} else {
-			cascadeOffset += titleHeight + borderWidth;
-		}
-
-	} else {
-		Gravitate(np, 0);
-	}
-
-	JXMoveWindow(display, np->parent, np->x - west, np->y - north);
-
-}
-
-/****************************************************************************
- * Get the offsets of the child window with respect to the parent.
- ****************************************************************************/
-void GetBorderOffsets(ClientNode *np, int *north, int *west) {
-
-	Assert(np);
-	Assert(north);
-	Assert(west);
-
-	*north = 0;
-	*west = 0;
-
-	if(np->borderFlags & BORDER_OUTLINE) {
-		*west = borderWidth;
-		if(np->borderFlags & BORDER_TITLE) {
-			*north = titleSize;
-		} else {
-			*north = borderWidth;
-		}
-	}
-
-}
-
-/****************************************************************************
- * Move the window in the specified direction for reparenting.
- ****************************************************************************/
-void Gravitate(ClientNode *np, int negate) {
-	int north, south, west;
-	int northDelta, westDelta;
-
-	Assert(np);
-
-	GetBorderOffsets(np, &north, &west);
-	if(north) {
-		south = borderWidth;
-	} else {
-		south = 0;
-	}
-
-	northDelta = 0;
-	westDelta = 0;
-	switch(np->gravity) {
-	case NorthWestGravity:
-		northDelta = -north;
-		westDelta = -west;
-		break;
-	case NorthGravity:
-		northDelta = -north;
-		break;
-	case NorthEastGravity:
-		northDelta = -north;
-		westDelta = west;
-		break;
-	case WestGravity:
-		westDelta = -west;
-		break;
-	case CenterGravity:
-		northDelta = (north + south) / 2;
-		westDelta = west;
-		break;
-	case EastGravity:
-		westDelta = west;
-		break;
-	case SouthWestGravity:
-		northDelta = south;
-		westDelta = -west;
-		break;
-	case SouthGravity:
-		northDelta = south;
-		break;
-	case SouthEastGravity:
-		northDelta = south;
-		westDelta = west;
-		break;
-	default: /* Static */
-		break;
-	}
-
-	if(negate) {
-		np->x += westDelta;
-		np->y += northDelta;
-	} else {
-		np->x -= westDelta;
-		np->y -= northDelta;
-	}
-
-}
-
 /****************************************************************************
  * Minimize a client window and all of its transients.
  ****************************************************************************/
@@ -404,11 +265,13 @@ void MinimizeClient(ClientNode *np) {
 	Assert(np);
 
 	if(focusModel == FOCUS_CLICK && np == activeClient) {
-		FocusNext();
+		FocusNextStacked(np);
 	}
 
 	MinimizeTransients(np);
-	DrawTray();
+
+	UpdateTaskBar();
+	UpdatePager();
 
 }
 
@@ -423,26 +286,26 @@ void MinimizeTransients(ClientNode *np) {
 
 	if(activeClient == np) {
 		activeClient = NULL;
-		np->statusFlags &= ~STAT_ACTIVE;
+		np->state.status &= ~STAT_ACTIVE;
 	}
 
-	if(np->statusFlags & STAT_SHADED) {
+	if(np->state.status & STAT_SHADED) {
 		UnshadeClient(np);
 	}
 
-	if(np->statusFlags & STAT_MAPPED) {
+	if(np->state.status & STAT_MAPPED) {
 		JXUnmapWindow(display, np->window);
 		JXUnmapWindow(display, np->parent);
 	}
-	np->statusFlags |= STAT_MINIMIZED;
-	np->statusFlags &= ~STAT_MAPPED;
-	np->statusFlags &= ~STAT_WITHDRAWN;
-	UpdateState(np);
+	np->state.status |= STAT_MINIMIZED;
+	np->state.status &= ~STAT_MAPPED;
+	np->state.status &= ~STAT_WITHDRAWN;
+	WriteState(np);
 
 	for(x = 0; x < LAYER_COUNT; x++) {
 		for(tp = nodes[x]; tp; tp = tp->next) {
-			if(tp->owner == np->window && (tp->statusFlags & STAT_MAPPED)
-				&& !(tp->statusFlags & STAT_MINIMIZED)) {
+			if(tp->owner == np->window && (tp->state.status & STAT_MAPPED)
+				&& !(tp->state.status & STAT_MINIMIZED)) {
 				MinimizeTransients(tp);
 			}
 		}
@@ -454,25 +317,31 @@ void MinimizeTransients(ClientNode *np) {
  ****************************************************************************/
 void ShadeClient(ClientNode *np) {
 
+	int bsize;
+
 	Assert(np);
 
-	if(!(np->borderFlags & BORDER_OUTLINE)
-		|| !(np->borderFlags & BORDER_TITLE)) {
+	if(!(np->state.border & BORDER_TITLE)) {
 		return;
 	}
+	if(np->state.border & BORDER_OUTLINE) {
+		bsize = borderWidth;
+	} else {
+		bsize = 0;
+	}
 
-	if(np->statusFlags & STAT_MAPPED) {
+	if(np->state.status & STAT_MAPPED) {
 		JXUnmapWindow(display, np->window);
 	}
-	np->statusFlags |= STAT_SHADED;
-	np->statusFlags &= ~STAT_MINIMIZED;
-	np->statusFlags &= ~STAT_MAPPED;
-	np->statusFlags &= ~STAT_WITHDRAWN;
+	np->state.status |= STAT_SHADED;
+	np->state.status &= ~STAT_MINIMIZED;
+	np->state.status &= ~STAT_MAPPED;
+	np->state.status &= ~STAT_WITHDRAWN;
 
-	JXResizeWindow(display, np->parent, np->width + 2 * borderWidth,
-		titleSize + borderWidth);
+	JXResizeWindow(display, np->parent, np->width + 2 * bsize,
+		titleHeight + 2 * bsize);
 
-	WriteWinState(np);
+	WriteState(np);
 
 }
 
@@ -481,24 +350,30 @@ void ShadeClient(ClientNode *np) {
  ****************************************************************************/
 void UnshadeClient(ClientNode *np) {
 
+	int bsize;
+
 	Assert(np);
 
-	if(!(np->borderFlags & BORDER_OUTLINE)
-		|| !(np->borderFlags & BORDER_TITLE)) {
+	if(!(np->state.border & BORDER_TITLE)) {
 		return;
 	}
-
-	if(np->statusFlags & STAT_SHADED) {
-		JXMapWindow(display, np->window);
-		np->statusFlags |= STAT_MAPPED;
-		np->statusFlags &= ~STAT_SHADED;
+	if(np->state.border & BORDER_OUTLINE) {
+		bsize = borderWidth;
+	} else {
+		bsize = 0;
 	}
-	np->statusFlags &= ~STAT_WITHDRAWN;
 
-	JXResizeWindow(display, np->parent, np->width + 2 * borderWidth,
-		np->height + titleSize + borderWidth);
+	if(np->state.status & STAT_SHADED) {
+		JXMapWindow(display, np->window);
+		np->state.status |= STAT_MAPPED;
+		np->state.status &= ~STAT_SHADED;
+	}
+	np->state.status &= ~STAT_WITHDRAWN;
 
-	WriteWinState(np);
+	JXResizeWindow(display, np->parent, np->width + 2 * bsize,
+		np->height + titleHeight + 2 * bsize);
+
+	WriteState(np);
 
 	RefocusClient();
 
@@ -515,81 +390,92 @@ void SetClientWithdrawn(ClientNode *np, int isWithdrawn) {
 
 		if(activeClient == np) {
 			activeClient = NULL;
-			np->statusFlags &= ~STAT_ACTIVE;
+			np->state.status &= ~STAT_ACTIVE;
 		}
-		np->statusFlags |= STAT_WITHDRAWN;
-		if(np->statusFlags & STAT_MAPPED) {
+		np->state.status |= STAT_WITHDRAWN;
+		if(np->state.status & STAT_MAPPED) {
 			JXUnmapWindow(display, np->window);
 			JXUnmapWindow(display, np->parent);
-			UpdateState(np);
+			WriteState(np);
 		}
 
 	} else {
-		np->statusFlags &= ~STAT_WITHDRAWN;
-		if(np->statusFlags & STAT_MAPPED) {
+		np->state.status &= ~STAT_WITHDRAWN;
+		if(np->state.status & STAT_MAPPED) {
 			JXMapWindow(display, np->window);
 			JXMapWindow(display, np->parent);
-			UpdateState(np);
+			WriteState(np);
 		}
 	}
 
-	DrawTray();
+	UpdateTaskBar();
+	UpdatePager();
 
 }
 
 /****************************************************************************
  ****************************************************************************/
-void RestoreSingleClient(ClientNode *np) {
+void RestoreTransients(ClientNode *np) {
 
-	Assert(np);
-
-	if(np->statusFlags & STAT_MINIMIZED) {
-		JXMapWindow(display, np->window);
-		JXMapWindow(display, np->parent);
-		np->statusFlags |= STAT_MAPPED;
-		np->statusFlags &= ~STAT_MINIMIZED;
-	}
-	RaiseClient(np);
-	UpdateState(np);
-	WriteWinState(np);
-}
-
-/****************************************************************************
- ****************************************************************************/
-void RestoreClient(ClientNode *np) {
 	ClientNode *tp;
 	int x;
 
 	Assert(np);
 
-	RestoreSingleClient(np);
+	if(!(np->state.status & STAT_MAPPED)) {
+		JXMapWindow(display, np->window);
+		JXMapWindow(display, np->parent);
+	}
+	np->state.status |= STAT_MAPPED;
+	np->state.status &= ~STAT_MINIMIZED;
+	np->state.status &= ~STAT_WITHDRAWN;
+
+	WriteState(np);
 
 	for(x = 0; x < LAYER_COUNT; x++) {
 		for(tp = nodes[x]; tp; tp = tp->next) {
-			if(tp->owner == np->window) {
-				RestoreSingleClient(tp);
+			if(tp->owner == np->window
+				&& !(tp->state.status & STAT_MAPPED)
+				&& (tp->state.status & STAT_MINIMIZED)) {
+				RestoreTransients(tp);
 			}
 		}
 	}
 
+	RaiseClient(np);
+
+}
+
+/****************************************************************************
+ ****************************************************************************/
+void RestoreClient(ClientNode *np) {
+
+	Assert(np);
+
+	RestoreTransients(np);
+
 	RestackClients();
+	UpdateTaskBar();
+	UpdatePager();
 
 }
 
 /****************************************************************************
  * Set the client layer. This will affect transients.
  ****************************************************************************/
-void SetClientLayer(ClientNode *np, int layer) {
+void SetClientLayer(ClientNode *np, unsigned int layer) {
+
 	ClientNode *tp, *next;
 	int x;
 
 	Assert(np);
 
-	if(layer < LAYER_BOTTOM || layer > LAYER_TOP) {
+	if(layer > LAYER_TOP) {
+		Warning("Client %s requested an invalid layer: %d", np->name, layer);
 		return;
 	}
 
-	if(np->layer != layer) {
+	if(np->state.layer != layer) {
 
 		/* Loop through all clients so we get transients. */
 		for(x = 0; x < LAYER_COUNT; x++) {
@@ -603,12 +489,12 @@ void SetClientLayer(ClientNode *np, int layer) {
 					if(next) {
 						next->prev = tp->prev;
 					} else {
-						nodeTail[tp->layer] = tp->prev;
+						nodeTail[tp->state.layer] = tp->prev;
 					}
 					if(tp->prev) {
 						tp->prev->next = next;
 					} else {
-						nodes[tp->layer] = next;
+						nodes[tp->state.layer] = next;
 					}
 
 					/* Insert into the new node list */
@@ -622,7 +508,7 @@ void SetClientLayer(ClientNode *np, int layer) {
 					nodes[layer] = tp;
 
 					/* Set the new layer */
-					tp->layer = layer;
+					tp->state.layer = layer;
 					SetCardinalAtom(tp->window, ATOM_WIN_LAYER, layer);
 
 					/* Make sure we continue on the correct layer list. */
@@ -650,7 +536,7 @@ void SetClientSticky(ClientNode *np, int isSticky) {
 
 	Assert(np);
 
-	if(np->statusFlags & STAT_STICKY) {
+	if(np->state.status & STAT_STICKY) {
 		old = 1;
 	} else {
 		old = 0;
@@ -660,9 +546,9 @@ void SetClientSticky(ClientNode *np, int isSticky) {
 		for(x = 0; x < LAYER_COUNT; x++) {
 			for(tp = nodes[x]; tp; tp = tp->next) {
 				if(tp == np || tp->owner == np->window) {
-					tp->statusFlags |= STAT_STICKY;
-					SetCardinalAtom(tp->window, ATOM_NET_WM_DESKTOP, 0xFFFFFFFF);
-					WriteWinState(np);
+					tp->state.status |= STAT_STICKY;
+					SetCardinalAtom(tp->window, ATOM_NET_WM_DESKTOP, ~0UL);
+					WriteState(tp);
 				}
 			}
 		}
@@ -670,8 +556,8 @@ void SetClientSticky(ClientNode *np, int isSticky) {
 		for(x = 0; x < LAYER_COUNT; x++) {
 			for(tp = nodes[x]; tp; tp = tp->next) {
 				if(tp == np || tp->owner == np->window) {
-					tp->statusFlags &= ~STAT_STICKY;
-					WriteWinState(tp);
+					tp->state.status &= ~STAT_STICKY;
+					WriteState(tp);
 				}
 			}
 		}
@@ -683,22 +569,22 @@ void SetClientSticky(ClientNode *np, int isSticky) {
 /****************************************************************************
  * Set a client's desktop. This will update transients.
  ****************************************************************************/
-void SetClientDesktop(ClientNode *np, int desktop) {
+void SetClientDesktop(ClientNode *np, unsigned int desktop) {
 	ClientNode *tp;
 	int x;
 
 	Assert(np);
 
-	if(desktop < 0 || desktop >= desktopCount) {
+	if(desktop >= desktopCount) {
 		return;
 	}
 
-	if(!(np->statusFlags & STAT_STICKY)) {
+	if(!(np->state.status & STAT_STICKY)) {
 		for(x = 0; x < LAYER_COUNT; x++) {
 			for(tp = nodes[x]; tp; tp = tp->next) {
 				if(tp == np || tp->owner == np->window) {
 
-					tp->desktop = desktop;
+					tp->state.desktop = desktop;
 
 					if(desktop == currentDesktop) {
 						ShowClient(tp);
@@ -706,11 +592,13 @@ void SetClientDesktop(ClientNode *np, int desktop) {
 						HideClient(tp);
 					}
 
-					SetCardinalAtom(tp->window, ATOM_NET_WM_DESKTOP, tp->desktop);
+					SetCardinalAtom(tp->window, ATOM_NET_WM_DESKTOP,
+						tp->state.desktop);
 				}
 			}
 		}
-		DrawTray();
+		UpdatePager();
+		UpdateTaskBar();
 	}
 
 }
@@ -725,8 +613,8 @@ void HideClient(ClientNode *np) {
 	if(activeClient == np) {
 		activeClient = NULL;
 	}
-	np->statusFlags |= STAT_HIDDEN;
-	if(np->statusFlags & (STAT_MAPPED | STAT_SHADED)) {
+	np->state.status |= STAT_HIDDEN;
+	if(np->state.status & (STAT_MAPPED | STAT_SHADED)) {
 		JXUnmapWindow(display, np->parent);
 	}
 }
@@ -738,11 +626,11 @@ void ShowClient(ClientNode *np) {
 
 	Assert(np);
 
-	if(np->statusFlags & STAT_HIDDEN) {
-		np->statusFlags &= ~STAT_HIDDEN;
-		if(np->statusFlags & (STAT_MAPPED | STAT_SHADED)) {
+	if(np->state.status & STAT_HIDDEN) {
+		np->state.status &= ~STAT_HIDDEN;
+		if(np->state.status & (STAT_MAPPED | STAT_SHADED)) {
 			JXMapWindow(display, np->parent);
-			if(np->statusFlags & STAT_ACTIVE) {
+			if(np->state.status & STAT_ACTIVE) {
 				FocusClient(np);
 			}
 		}
@@ -752,67 +640,34 @@ void ShowClient(ClientNode *np) {
 /****************************************************************************
  ****************************************************************************/
 void MaximizeClient(ClientNode *np) {
+
 	int north, west;
 
 	Assert(np);
 
-	if(np->statusFlags & STAT_SHADED) {
+	if(np->state.status & STAT_SHADED) {
 		UnshadeClient(np);
 	}
 
-	if(np->borderFlags & BORDER_OUTLINE) {
+	if(np->state.border & BORDER_OUTLINE) {
 		west = borderWidth;
 	} else {
 		west = 0;
 	}
-	if(np->borderFlags & BORDER_TITLE) {
-		north = titleSize;
+	if(np->state.border & BORDER_TITLE) {
+		north = west + titleHeight;
 	} else {
 		north = west;
 	}
 
-	if(np->statusFlags & STAT_MAXIMIZED) {
+	if(np->state.status & STAT_MAXIMIZED) {
 		np->x = np->oldx;
 		np->y = np->oldy;
 		np->width = np->oldWidth;
 		np->height = np->oldHeight;
-		np->statusFlags &= ~STAT_MAXIMIZED;
+		np->state.status &= ~STAT_MAXIMIZED;
 	} else {
-		np->oldx = np->x;
-		np->oldy = np->y;
-		np->oldWidth = np->width;
-		np->oldHeight = np->height;
-
-		np->x = west;
-		np->y = north;
-		np->width = rootWidth - 2 * borderWidth;
-		np->height = rootHeight - titleSize - borderWidth;
-		if(!autoHideTray) {
-			np->height -= trayHeight;
-		}
-
-		if(np->width > np->maxWidth) {
-			np->width = np->maxWidth;
-		}
-		if(np->height > np->maxHeight) {
-			np->height = np->maxHeight;
-		}
-
-		if(np->sizeFlags & PAspect) {
-			if((float)np->width / np->height
-				< (float)np->aspect.minx / np->aspect.miny) {
-				np->height = np->width * np->aspect.miny / np->aspect.minx;
-			}
-			if((float)np->width / np->height
-				> (float)np->aspect.maxx / np->aspect.maxy) {
-				np->width = np->height * np->aspect.maxx / np->aspect.maxy;
-			}
-		}
-
-		np->width -= np->width % np->xinc;
-		np->height -= np->height % np->yinc;
-
-		np->statusFlags |= STAT_MAXIMIZED;
+		PlaceMaximizedClient(np);
 	}
 
 	JXMoveResizeWindow(display, np->parent,
@@ -822,8 +677,8 @@ void MaximizeClient(ClientNode *np) {
 	JXMoveResizeWindow(display, np->window, west,
 		north, np->width, np->height);
 
+	WriteState(np);
 	SendConfigureEvent(np);
-	WriteWinState(np);
 
 }
 
@@ -833,32 +688,34 @@ void FocusClient(ClientNode *np) {
 
 	Assert(np);
 
-	if(np->statusFlags & (STAT_MINIMIZED | STAT_WITHDRAWN | STAT_HIDDEN)) {
+	if(np->state.status & (STAT_MINIMIZED | STAT_WITHDRAWN | STAT_HIDDEN)) {
 		return;
 	}
 
 	if(activeClient != np) {
 		if(activeClient) {
-			activeClient->statusFlags &= ~STAT_ACTIVE;
+			activeClient->state.status &= ~STAT_ACTIVE;
 			DrawBorder(activeClient);
 		}
-		np->statusFlags |= STAT_ACTIVE;
+		np->state.status |= STAT_ACTIVE;
 		activeClient = np;
 
-		if(!(np->statusFlags & STAT_SHADED)) {
+		if(!(np->state.status & STAT_SHADED)) {
 			UpdateClientColormap(np);
 			if(np->protocols & PROT_TAKE_FOCUS) {
-				SendClientMessage(np, ATOM_WM_PROTOCOLS, ATOM_WM_TAKE_FOCUS);
+				SendClientMessage(np->window, ATOM_WM_PROTOCOLS,
+					ATOM_WM_TAKE_FOCUS);
 			}
 			SetWindowAtom(rootWindow, ATOM_NET_ACTIVE_WINDOW, np->window);
 		}
 
 		DrawBorder(np);
-		DrawTray();
+		UpdatePager();
+		UpdateTaskBar();
 
 	}
 
-	if(np->statusFlags & STAT_MAPPED && !(np->statusFlags & STAT_HIDDEN)) {
+	if(np->state.status & STAT_MAPPED && !(np->state.status & STAT_HIDDEN)) {
 		JXSetInputFocus(display, np->window, RevertToPointerRoot, CurrentTime);
 	} else {
 		JXSetInputFocus(display, rootWindow, RevertToPointerRoot, CurrentTime);
@@ -866,11 +723,37 @@ void FocusClient(ClientNode *np) {
 
 }
 
+/****************************************************************************
+ ****************************************************************************/
+void FocusNextStacked(ClientNode *np) {
+
+	const int FLAGS = STAT_MINIMIZED | STAT_WITHDRAWN | STAT_HIDDEN;
+	int x;
+	ClientNode *tp;
+
+	Assert(np);
+
+	for(tp = np->next; tp; tp = tp->next) {
+		if(!(tp->state.status & FLAGS)) {
+			FocusClient(tp);
+			return;
+		}
+	}
+	for(x = np->state.layer - 1; x >= LAYER_BOTTOM; x--) {
+		for(tp = nodes[x]; tp; tp = tp->next) {
+			if(!(tp->state.status & FLAGS)) {
+				FocusClient(tp);
+				return;
+			}
+		}
+	}
+
+}
+
 /****************************************************************************
  * Refocus the active client, if existent.
  ****************************************************************************/
 void RefocusClient() {
-
 	if(activeClient) {
 		FocusClient(activeClient);
 	}
@@ -884,10 +767,12 @@ void DeleteClient(ClientNode *np) {
 
 	ReadWMProtocols(np);
 	if(np->protocols & PROT_DELETE) {
-		SendClientMessage(np, ATOM_WM_PROTOCOLS, ATOM_WM_DELETE_WINDOW);
+		SendClientMessage(np->window, ATOM_WM_PROTOCOLS,
+			ATOM_WM_DELETE_WINDOW);
 	} else {
 		KillClient(np);
 	}
+
 }
 
 /****************************************************************************
@@ -897,10 +782,20 @@ void KillClientHandler(ClientNode *np) {
 	Assert(np);
 
 	if(np == activeClient) {
-		activeClient = NULL;
+		FocusNextStacked(np);
 	}
 
+	JXGrabServer(display);
+	JXSync(display, False);
+
 	JXKillClient(display, np->window);
+
+	JXSync(display, True);
+	JXUngrabServer(display);
+
+	np->window = None;
+	RemoveClient(np);
+
 }
 
 /****************************************************************************
@@ -924,7 +819,7 @@ void RaiseClient(ClientNode *np) {
 
 	Assert(np);
 
-	if(np->layer > LAYER_BOTTOM && nodes[np->layer] != np) {
+	if(np->state.layer > LAYER_BOTTOM && nodes[np->state.layer] != np) {
 
 		/* Raise the window */
 		Assert(np->prev);
@@ -932,12 +827,12 @@ void RaiseClient(ClientNode *np) {
 		if(np->next) {
 			np->next->prev = np->prev;
 		} else {
-			nodeTail[np->layer] = np->prev;
+			nodeTail[np->state.layer] = np->prev;
 		}
-		np->next = nodes[np->layer];
-		nodes[np->layer]->prev = np;
+		np->next = nodes[np->state.layer];
+		nodes[np->state.layer]->prev = np;
 		np->prev = NULL;
-		nodes[np->layer] = np;
+		nodes[np->state.layer] = np;
 
 		/* Place any transient windows on top of the owner */
 		for(x = 0; x < LAYER_COUNT; x++) {
@@ -950,12 +845,12 @@ void RaiseClient(ClientNode *np) {
 					if(tp->next) {
 						tp->next->prev = tp->prev;
 					} else {
-						nodeTail[tp->layer] = tp->prev;
+						nodeTail[tp->state.layer] = tp->prev;
 					}
-					tp->next = nodes[tp->layer];
-					nodes[tp->layer]->prev = tp;
+					tp->next = nodes[tp->state.layer];
+					nodes[tp->state.layer]->prev = tp;
 					tp->prev = NULL;
-					nodes[tp->layer] = tp;
+					nodes[tp->state.layer] = tp;
 
 					tp = next;
 
@@ -976,58 +871,79 @@ void RaiseClient(ClientNode *np) {
 /****************************************************************************
  ****************************************************************************/
 void RestackClients() {
+
+	TrayType *tp;
 	ClientNode *np;
-	int layer, index;
+	unsigned int layer, index;
+	int trayCount;
 
 	if(!stack) {
 		return;
 	}
 
-	/* Make sure an appropriate amount of memory is allocated. */
-	if(clientCount + 1 >= stackSize
-		|| clientCount + 1 < stackSize - STACK_BLOCK_SIZE) {
+	/* Determine how many tray windows exist. */
+	trayCount = 0;
+	for(tp = GetTrays(); tp; tp = tp->next) {
+		++trayCount;
+	}
 
-		stackSize = (clientCount + 1) + STACK_BLOCK_SIZE
-			- ((clientCount + 1) % STACK_BLOCK_SIZE);
+	/* Make sure an appropriate amount of memory is allocated. */
+	if(clientCount + trayCount >= stackSize
+		|| clientCount + trayCount < stackSize - STACK_BLOCK_SIZE) {
+
+		stackSize = (clientCount + trayCount) + STACK_BLOCK_SIZE
+			- ((clientCount + trayCount) % STACK_BLOCK_SIZE);
 		stack = Reallocate(stack, sizeof(Window) * stackSize);
 
 	}
 
 	/* Prepare the stacking array. */
 	index = 0;
-	for(layer = LAYER_TOP; layer >= LAYER_BOTTOM; layer--) {
+	layer = LAYER_TOP;
+	for(;;) {
+
 		for(np = nodes[layer]; np; np = np->next) {
-			if(!(np->statusFlags
+			if(!(np->state.status
 				& (STAT_MINIMIZED | STAT_HIDDEN | STAT_WITHDRAWN))) {
 				stack[index++] = np->parent;
 			}
 		}
-		if(layer == LAYER_TRAY) {
-			stack[index++] = trayWindow;
+
+		for(tp = GetTrays(); tp; tp = tp->next) {
+			if(layer == tp->layer) {
+				stack[index++] = tp->window;
+			}
 		}
+
+		if(layer == 0) {
+			break;
+		}
+		--layer;
+
 	}
 
 	JXRestackWindows(display, stack, index);
 
+	UpdateNetClientList();
+
 }
 
 /****************************************************************************
  ****************************************************************************/
-void SendClientMessage(ClientNode *np, AtomType type, AtomType message) {
+void SendClientMessage(Window w, AtomType type, AtomType message) {
+
 	XEvent event;
 	int status;
 
-	Assert(np);
-
 	memset(&event, 0, sizeof(event));
 	event.xclient.type = ClientMessage;
-	event.xclient.window = np->window;
+	event.xclient.window = w;
 	event.xclient.message_type = atoms[type];
 	event.xclient.format = 32;
 	event.xclient.data.l[0] = atoms[message];
 	event.xclient.data.l[1] = CurrentTime;
 
-	status = JXSendEvent(display, np->window, False, 0, &event);
+	status = JXSendEvent(display, w, False, 0, &event);
 	if(status == False) {
 		Debug("SendClientMessage failed");
 	}
@@ -1037,18 +953,19 @@ void SendClientMessage(ClientNode *np, AtomType type, AtomType message) {
  ****************************************************************************/
 #ifdef USE_SHAPE
 void SetShape(ClientNode *np) {
+
 	XRectangle rect[4];
 	int north, west;
 
 	Assert(np);
 
-	if(np->borderFlags & BORDER_OUTLINE) {
+	if(np->state.border & BORDER_OUTLINE) {
 		west = borderWidth;
 	} else {
 		west = 0;
 	}
-	if(np->borderFlags & BORDER_TITLE) {
-		north = titleSize;
+	if(np->state.border & BORDER_TITLE) {
+		north = west + titleHeight;
 	} else {
 		north = west;
 	}
@@ -1092,26 +1009,32 @@ void SetShape(ClientNode *np) {
 /****************************************************************************
  ****************************************************************************/
 void RemoveClient(ClientNode *np) {
+
 	ColormapNode *cp;
 
 	Assert(np);
 
+	JXGrabServer(display);
+
 	/* Remove this client from the client list */
 	if(np->next) {
 		np->next->prev = np->prev;
 	} else {
-		nodeTail[np->layer] = np->prev;
+		nodeTail[np->state.layer] = np->prev;
 	}
 	if(np->prev) {
 		np->prev->next = np->next;
 	} else {
-		nodes[np->layer] = np->next;
+		nodes[np->state.layer] = np->next;
 	}
 	--clientCount;
 	XDeleteContext(display, np->window, clientContext);
 	XDeleteContext(display, np->parent, frameContext);
 
 	/* Make sure this client isn't active */
+	if(activeClient == np && !shouldExit) {
+		FocusNextStacked(np);
+	}
 	if(activeClient == np) {
 		SetWindowAtom(rootWindow, ATOM_NET_ACTIVE_WINDOW, None);
 		activeClient = NULL;
@@ -1119,10 +1042,18 @@ void RemoveClient(ClientNode *np) {
 
 	/* If the window manager is exiting (ie, not the client), then
 	 * reparent etc. */
-	if(shouldExit && !(np->statusFlags & STAT_WMDIALOG)) {
-		Gravitate(np, 1);
-		if(!(np->statusFlags & STAT_MAPPED)
-			&& (np->statusFlags & (STAT_MINIMIZED | STAT_SHADED))) {
+	if(shouldExit && !(np->state.status & STAT_WMDIALOG)) {
+		if(np->state.status & STAT_MAXIMIZED) {
+			np->x = np->oldx;
+			np->y = np->oldy;
+			np->width = np->oldWidth;
+			np->height = np->oldHeight;
+			JXMoveResizeWindow(display, np->window,
+				np->x, np->y, np->width, np->height);
+		}
+		GravitateClient(np, 1);
+		if(!(np->state.status & STAT_MAPPED)
+			&& (np->state.status & (STAT_MINIMIZED | STAT_SHADED))) {
 			JXMapWindow(display, np->window);
 		}
 		JXUngrabButton(display, AnyButton, AnyModifier, np->window);
@@ -1146,7 +1077,9 @@ void RemoveClient(ClientNode *np) {
 		JXFree(np->className);
 	}
 
-	RemoveClientFromTray(np);
+	RemoveClientFromTaskBar(np);
+	RemoveClientStrut(np);
+	UpdatePager();
 
 	while(np->colormaps) {
 		cp = np->colormaps->next;
@@ -1158,6 +1091,8 @@ void RemoveClient(ClientNode *np) {
 
 	Release(np);
 
+	JXUngrabServer(display);
+
 }
 
 /****************************************************************************
@@ -1194,7 +1129,9 @@ ClientNode *FindClientByParent(Window p) {
 /****************************************************************************
  ****************************************************************************/
 void ReparentClient(ClientNode *np, int notOwner) {
+
 	XSetWindowAttributes attr;
+	int attrMask;
 	int x, y, width, height;
 
 	Assert(np);
@@ -1213,43 +1150,65 @@ void ReparentClient(ClientNode *np, int notOwner) {
 		True, ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);
 	GrabKeys(np);
 
+	attrMask = 0;
+
+	attrMask |= CWOverrideRedirect;
 	attr.override_redirect = True;
-	attr.event_mask = ButtonPressMask | ButtonReleaseMask | ExposureMask
-		| PointerMotionMask | SubstructureRedirectMask | SubstructureNotifyMask
-		| EnterWindowMask | LeaveWindowMask | KeyPressMask;
+
+	attrMask |= CWEventMask;
+	attr.event_mask
+		= ButtonPressMask
+		| ButtonReleaseMask
+		| ExposureMask
+		| PointerMotionMask
+		| SubstructureRedirectMask
+		| SubstructureNotifyMask
+		| EnterWindowMask
+		| LeaveWindowMask
+		| KeyPressMask;
+
+	attrMask |= CWDontPropagate;
+	attr.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
+
+	attrMask |= CWBackPixel;
 	attr.background_pixel = colors[COLOR_BORDER_BG];
 
-	if(np->borderFlags & BORDER_OUTLINE) {
-		x = np->x - borderWidth;
-		width = np->width + borderWidth + borderWidth;
-		if(np->borderFlags & BORDER_TITLE) {
-			y = np->y - titleSize;
-			height = np->height + titleSize + borderWidth;
-		} else {
-			y = np->y - borderWidth;
-			height = np->height + 2 * borderWidth;
-		}
-	} else {
-		x = np->x;
-		y = np->y;
-		width = np->width;
-		height = np->height;
+	x = np->x;
+	y = np->y;
+	width = np->width;
+	height = np->height;
+	if(np->state.border & BORDER_OUTLINE) {
+		x -= borderWidth;
+		y -= borderWidth;
+		width += borderWidth * 2;
+		height += borderWidth * 2;
+	}
+	if(np->state.border & BORDER_TITLE) {
+		y -= titleHeight;
+		height += titleHeight;
 	}
 
 	np->parent = JXCreateWindow(display, rootWindow,
 		x, y, width, height, 0, rootDepth, InputOutput,
-		rootVisual, CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
+		rootVisual, attrMask, &attr);
 	np->parentGC = JXCreateGC(display, np->parent, 0, NULL);
 
+	attrMask = CWDontPropagate;
+	attr.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
+
+	JXChangeWindowAttributes(display, np->window, attrMask, &attr);
+
 	JXSetWindowBorderWidth(display, np->window, 0);
-	if(np->borderFlags & BORDER_OUTLINE) {
-		if(np->borderFlags & BORDER_TITLE) {
-			JXReparentWindow(display, np->window, np->parent, borderWidth,
-				titleSize);
-		} else {
-			JXReparentWindow(display, np->window, np->parent, borderWidth,
-				borderWidth);
-		}
+	if((np->state.border & BORDER_OUTLINE)
+		&& (np->state.border & BORDER_TITLE)) {
+		JXReparentWindow(display, np->window, np->parent,
+			borderWidth, borderWidth + titleHeight);
+	} else if(np->state.border & BORDER_OUTLINE) {
+		JXReparentWindow(display, np->window, np->parent,
+			borderWidth, borderWidth);
+	} else if(np->state.border & BORDER_TITLE) {
+		JXReparentWindow(display, np->window, np->parent,
+			0, titleHeight);
 	} else {
 		JXReparentWindow(display, np->window, np->parent, 0, 0);
 	}
@@ -1277,7 +1236,7 @@ void CheckShape(ClientNode *np) {
 		&xb, &yb, &wb, &hb, &clipShaped, &xc, &yc, &wc, &hc);
 
 	if(boundingShaped == True) {
-		np->statusFlags |= STAT_USESHAPE;
+		np->state.status |= STAT_USESHAPE;
 		SetShape(np);
 	}
 }
@@ -1311,12 +1270,15 @@ void SendConfigureEvent(ClientNode *np) {
  * client is active.
  ****************************************************************************/
 void UpdateClientColormap(ClientNode *np) {
+
 	XWindowAttributes attr;
 	ColormapNode *cp;
 	int wasInstalled;
 
 	Assert(np);
 
+	cp = np->colormaps;
+
 	if(np == activeClient) {
 
 		wasInstalled = 0;
@@ -1341,87 +1303,3 @@ void UpdateClientColormap(ClientNode *np) {
 
 }
 
-/****************************************************************************
- ****************************************************************************/
-void UpdateState(ClientNode *np) {
-	unsigned long data[2];
-
-	Assert(np);
-
-	if(np->statusFlags & STAT_WITHDRAWN) {
-		data[0] = WithdrawnState;
-	} else if(np->statusFlags & STAT_MINIMIZED) {
-		data[0] = IconicState;
-	} else {
-		data[0] = NormalState;
-	}
-	data[1] = None;
-
-	JXChangeProperty(display, np->window, atoms[ATOM_WM_STATE],
-		atoms[ATOM_WM_STATE], 32, PropModeReplace,
-		(unsigned char*)data, 2);
-
-}
-
-/****************************************************************************
- * Read Motif hints.
- ****************************************************************************/
-void ReadMotifHints(ClientNode *np) {
-	PropMwmHints *mhints;
-	Atom type;
-	unsigned long itemCount, bytesLeft;
-	unsigned char *data;
-	int format;
-
-	Assert(np);
-
-	if(JXGetWindowProperty(display, np->window, atoms[ATOM_MOTIF_WM_HINTS],
-		0L, 20L, False, atoms[ATOM_MOTIF_WM_HINTS], &type, &format,
-		&itemCount, &bytesLeft, &data) != Success) {
-		return;
-	}
-
-	mhints = (PropMwmHints*)data;
-	if(mhints) {
-
-		if((mhints->flags & MWM_HINTS_FUNCTIONS)
-			&& !(mhints->functions & MWM_FUNC_ALL)) {
-
-			if(!(mhints->functions & MWM_FUNC_RESIZE)) {
-				np->borderFlags &= ~BORDER_RESIZE;
-			}
-			if(!(mhints->functions & MWM_FUNC_MOVE)) {
-				np->borderFlags &= ~BORDER_MOVE;
-			}
-			if(!(mhints->functions & MWM_FUNC_MINIMIZE)) {
-				np->borderFlags &= ~BORDER_MIN;
-			}
-			if(!(mhints->functions & MWM_FUNC_MAXIMIZE)) {
-				np->borderFlags &= ~BORDER_MAX;
-			}
-			if(!(mhints->functions & MWM_FUNC_CLOSE)) {
-				np->borderFlags &= ~BORDER_CLOSE;
-			}
-		}
-
-		if((mhints->flags & MWM_HINTS_DECORATIONS)
-			&& !(mhints->decorations & MWM_DECOR_ALL)) {
-
-			if(!(mhints->decorations & MWM_DECOR_BORDER)) {
-				np->borderFlags &= ~BORDER_OUTLINE;
-			}
-			if(!(mhints->decorations & MWM_DECOR_TITLE)) {
-				np->borderFlags &= ~BORDER_TITLE;
-			}
-			if(!(mhints->decorations & MWM_DECOR_MINIMIZE)) {
-				np->borderFlags &= ~BORDER_MIN;
-			}
-			if(!(mhints->decorations & MWM_DECOR_MAXIMIZE)) {
-				np->borderFlags &= ~BORDER_MAX;
-			}
-		}
-
-		JXFree(mhints);
-	}
-}
-
