@@ -6,6 +6,8 @@
 #ifndef CURSOR_H
 #define CURSOR_H
 
+#include "border.h"
+
 void InitializeCursors();
 void StartupCursors();
 void ShutdownCursors();
@@ -13,11 +15,12 @@ void DestroyCursors();
 
 int GrabMouseForMove();
 int GrabMouseForMenu();
-int GrabMouseForWM(Window w);
 int GrabMouseForResize(BorderActionType action);
 
 Cursor GetFrameCursor(BorderActionType action);
 
+void SetMousePosition(Window win, int x, int y);
+
 void GetMousePosition(int *x, int *y);
 unsigned int GetMouseMask();
 
