@@ -4,17 +4,43 @@
  ****************************************************************************/
 
 #include "jwm.h"
+#include "event.h"
+#include "client.h"
+#include "main.h"
+#include "hint.h"
+#include "tray.h"
+#include "pager.h"
+#include "desktop.h"
+#include "cursor.h"
+#include "icon.h"
+#include "taskbar.h"
+#include "confirm.h"
+#include "swallow.h"
+#include "popup.h"
+#include "winmenu.h"
+#include "root.h"
+#include "move.h"
+#include "resize.h"
+#include "key.h"
+#include "clock.h"
+#include "place.h"
+#include "dock.h"
+#include "timing.h"
+#include "traybutton.h"
 
+static unsigned int onRootMask = ~0;
+
+static void Signal();
 static void DispatchBorderButtonEvent(const XButtonEvent *event,
 	ClientNode *np);
 
 static void HandleConfigureRequest(const XConfigureRequestEvent *event);
 static int HandleExpose(const XExposeEvent *event);
 static int HandlePropertyNotify(const XPropertyEvent *event);
-static void HandleConfigureNotify(const XConfigureEvent *event);
+static int HandleConfigureNotify(const XConfigureEvent *event);
 static void HandleClientMessage(const XClientMessageEvent *event);
 static void HandleColormapChange(const XColormapEvent *event);
-static void HandleDestroyNotify(const XDestroyWindowEvent *event);
+static int HandleDestroyNotify(const XDestroyWindowEvent *event);
 static void HandleMapRequest(const XMapEvent *event);
 static void HandleUnmapNotify(const XUnmapEvent *event);
 static void HandleButtonEvent(const XButtonEvent *event);
@@ -22,18 +48,60 @@ static void HandleKeyPress(const XKeyEvent *event);
 static void HandleEnterNotify(const XCrossingEvent *event);
 static void HandleLeaveNotify(const XCrossingEvent *event);
 static void HandleMotionNotify(const XMotionEvent *event);
+static int HandleSelectionClear(const XSelectionClearEvent *event);
+
+static void HandleNetMoveResize(const XClientMessageEvent *event,
+	ClientNode *np);
+static void HandleNetWMState(const XClientMessageEvent *event,
+	ClientNode *np);
 
 #ifdef USE_SHAPE
 static void HandleShapeEvent(const XShapeEvent *event);
 #endif
 
+/****************************************************************************
+ * Handle events that come through while JWM is starting.
+ ****************************************************************************/
+void HandleStartupEvents() {
+
+	const long eventMask = SubstructureRedirectMask;
+
+	XEvent event;
+	XWindowChanges wc;
+
+	while(JXCheckMaskEvent(display, eventMask, &event)) {
+
+		switch(event.type) {
+		case ConfigureRequest:
+			wc.stack_mode = event.xconfigurerequest.detail;
+			wc.sibling = event.xconfigurerequest.above;
+			wc.border_width = event.xconfigurerequest.border_width;
+			wc.x = event.xconfigurerequest.x;
+			wc.y = event.xconfigurerequest.y;
+			wc.width = event.xconfigurerequest.width;
+			wc.height = event.xconfigurerequest.height;
+			JXConfigureWindow(display, event.xconfigurerequest.window,
+				event.xconfigurerequest.value_mask, &wc);
+			break;
+		case MapRequest:
+			JXMapWindow(display, event.xmap.window);
+			break;
+		default:
+			break;
+		}
+
+	}
+
+}
+
 /****************************************************************************
  ****************************************************************************/
 void WaitForEvent(XEvent *event) {
+
 	struct timeval timeout;
+	fd_set fds;
 	int fd;
 	int handled;
-	fd_set fds;
 
 	fd = JXConnectionNumber(display);
 
@@ -45,11 +113,11 @@ void WaitForEvent(XEvent *event) {
 			timeout.tv_usec = 0;
 			timeout.tv_sec = 1;
 			if(select(fd + 1, &fds, NULL, NULL, &timeout) <= 0) {
-				UpdateTime();
+				Signal();
 			}
 		}
 
-		UpdateTime();
+		Signal();
 
 		JXNextEvent(display, event);
 
@@ -81,12 +149,16 @@ void WaitForEvent(XEvent *event) {
 			handled = 1;
 			break;
 		case DestroyNotify:
-			HandleDestroyNotify(&event->xdestroywindow);
-			handled = 1;
+			handled = HandleDestroyNotify(&event->xdestroywindow);
 			break;
 		case ConfigureNotify:
-			HandleConfigureNotify(&event->xconfigure);
-			handled = 1;
+			handled = HandleConfigureNotify(&event->xconfigure);
+			break;
+		case SelectionClear:
+			handled = HandleSelectionClear(&event->xselectionclear);
+			break;
+		case ResizeRequest:
+			handled = HandleDockResizeRequest(&event->xresizerequest);
 			break;
 		case CreateNotify:
 		case MapNotify:
@@ -98,7 +170,7 @@ void WaitForEvent(XEvent *event) {
 		default:
 #ifdef USE_SHAPE
 			if(haveShape && event->type == shapeEvent) {
-				HandleShapeEvent((XShapeEvent*)(void*)&event);
+				HandleShapeEvent((XShapeEvent*)event);
 				handled = 1;
 			} else {
 				handled = 0;
@@ -115,10 +187,30 @@ void WaitForEvent(XEvent *event) {
 		if(!handled) {
 			handled = ProcessDialogEvent(event);
 		}
+		if(!handled) {
+			handled = ProcessSwallowEvent(event);
+		}
 
 		handled |= ProcessPopupEvent(event);
 
-	} while(handled);
+	} while(handled && !shouldExit);
+
+}
+
+/****************************************************************************
+ ****************************************************************************/
+void Signal() {
+
+	TimeType now;
+	int x, y;
+
+	GetCurrentTime(&now);
+	GetMousePosition(&x, &y);
+
+	SignalTaskbar(&now, x, y);
+	SignalTrayButton(&now, x, y);
+	SignalClock(&now, x, y);
+	SignalTray(&now, x, y);
 
 }
 
@@ -144,7 +236,10 @@ void ProcessEvent(XEvent *event) {
 		while(JXCheckTypedEvent(display, MotionNotify, event));
 		HandleMotionNotify(&event->xmotion);
 		break;
+	case DestroyNotify:
+	case Expose:
 	case KeyRelease:
+	case ConfigureNotify:
 		break;
 	default:
 		Debug("Unknown event type: %d", event->type);
@@ -152,6 +247,14 @@ void ProcessEvent(XEvent *event) {
 	}
 }
 
+/****************************************************************************
+ ****************************************************************************/
+int HandleSelectionClear(const XSelectionClearEvent *event) {
+
+	return HandleDockSelectionClear(event);
+
+}
+
 /****************************************************************************
  ****************************************************************************/
 void HandleButtonEvent(const XButtonEvent *event) {
@@ -174,17 +277,12 @@ void HandleButtonEvent(const XButtonEvent *event) {
 		case Button3:
 			x = event->x + np->x;
 			y = event->y + np->y;
-			if(np->borderFlags & BORDER_OUTLINE) {
+			if(np->state.border & BORDER_OUTLINE) {
 				x -= borderWidth;
-				if(np->borderFlags & BORDER_TITLE) {
-					y -= titleSize;
-				} else {
-					y -= borderWidth;
-				}
-			} else {
-				if(np->borderFlags & BORDER_TITLE) {
-					y -= titleSize;
-				}
+				y -= borderWidth;
+			}
+			if(np->state.border & BORDER_TITLE) {
+				y -= titleHeight;
 			}
 			ShowWindowMenu(np, x, y);
 			break;
@@ -192,7 +290,9 @@ void HandleButtonEvent(const XButtonEvent *event) {
 			break;
 		}
 	} else if(event->window == rootWindow) {
-		ShowRootMenu(event->x, event->y);
+		if((1 << event->button) & onRootMask) {
+			ShowRootMenu(event->x, event->y);
+		}
 	} else {
 		np = FindClientByWindow(event->window);
 		if(np) {
@@ -223,11 +323,7 @@ void HandleKeyPress(const XKeyEvent *event) {
 
 	key = GetKey(event);
 
-	if(focusModel == FOCUS_CLICK) {
-		np = FindClientByWindow(event->subwindow);
-	} else {
-		np = GetActiveClient();
-	}
+	np = GetActiveClient();
 
 	switch(key & 0xFF) {
 	case KEY_EXEC:
@@ -243,6 +339,9 @@ void HandleKeyPress(const XKeyEvent *event) {
 	case KEY_NEXT:
 		FocusNext();
 		break;
+	case KEY_NEXT_STACKED:
+		FocusNextStackedCircular();
+		break;
 	case KEY_CLOSE:
 		if(np) {
 			DeleteClient(np);
@@ -250,7 +349,7 @@ void HandleKeyPress(const XKeyEvent *event) {
 		break;
 	case KEY_SHADE:
 		if(np) {
-			if(np->statusFlags & STAT_SHADED) {
+			if(np->state.status & STAT_SHADED) {
 				UnshadeClient(np);
 			} else {
 				ShadeClient(np);
@@ -300,6 +399,7 @@ void HandleKeyPress(const XKeyEvent *event) {
 /****************************************************************************
  ****************************************************************************/
 void HandleConfigureRequest(const XConfigureRequestEvent *event) {
+
 	XWindowChanges wc;
 	ClientNode *np;
 	int north, south, east, west;
@@ -334,26 +434,26 @@ void HandleConfigureRequest(const XConfigureRequestEvent *event) {
 			return;
 		}
 
-		if(np->borderFlags & BORDER_OUTLINE) {
-			east = borderWidth;
-			west = borderWidth;
-			south = borderWidth;
-			if(np->borderFlags & BORDER_TITLE) {
-				north = titleSize;
-			} else {
-				north = borderWidth;
-			}
-		} else {
-			north = 0;
-			south = 0;
-			east = 0;
-			west = 0;
+		north = 0;
+		south = 0;
+		east = 0;
+		west = 0;
+		if(np->state.border & BORDER_OUTLINE) {
+			north += borderWidth;
+			south += borderWidth;
+			east += borderWidth;
+			west += borderWidth;
+		}
+		if(np->state.border & BORDER_TITLE) {
+			north += titleHeight;
 		}
 
 		wc.stack_mode = Above;
 		wc.sibling = np->parent;
 		wc.border_width = 0;
 
+		ConstrainSize(np);
+
 		wc.x = np->x;
 		wc.y = np->y;
 		wc.width = np->width + east + west;
@@ -367,14 +467,16 @@ void HandleConfigureRequest(const XConfigureRequestEvent *event) {
 		JXConfigureWindow(display, np->window, event->value_mask, &wc);
 
 	} else {
+
 		wc.stack_mode = event->detail;
 		wc.sibling = event->above;
 		wc.border_width = event->border_width;
 		wc.x = event->x;
 		wc.y = event->y;
-		wc.width = event->width;
-		wc.height = event->height;
+		wc.width = event->width > rootWidth ? rootWidth : event->width;
+		wc.height = event->height > rootHeight ? rootHeight : event->height;
 		JXConfigureWindow(display, event->window, event->value_mask, &wc);
+
 	}
 
 }
@@ -387,7 +489,7 @@ void HandleEnterNotify(const XCrossingEvent *event) {
 
 	np = FindClientByWindow(event->window);
 	if(np) {
-		if(!(np->statusFlags & STAT_ACTIVE) && (focusModel == FOCUS_SLOPPY)) {
+		if(!(np->state.status & STAT_ACTIVE) && (focusModel == FOCUS_SLOPPY)) {
 			FocusClient(np);
 		}
 		if(np->parent == event->window) {
@@ -427,7 +529,7 @@ int HandleExpose(const XExposeEvent *event) {
 			DrawBorder(np);
 			return 1;
 		} else if(event->window == np->window
-			&& np->statusFlags & STAT_WMDIALOG) {
+			&& np->state.status & STAT_WMDIALOG) {
 			return 0;
 		} else {
 			return 1;
@@ -440,6 +542,7 @@ int HandleExpose(const XExposeEvent *event) {
 /****************************************************************************
  ****************************************************************************/
 int HandlePropertyNotify(const XPropertyEvent *event) {
+
 	ClientNode *np;
 	int changed;
 
@@ -466,15 +569,23 @@ int HandlePropertyNotify(const XPropertyEvent *event) {
 			} else if(event->atom == atoms[ATOM_NET_WM_ICON]) {
 				LoadIcon(np);
 				changed = 1;
+			} else if(event->atom == atoms[ATOM_NET_WM_NAME]) {
+				ReadWMName(np);
+				changed = 1;
+			} else if(event->atom == atoms[ATOM_NET_WM_STRUT_PARTIAL]) {
+				ReadClientStrut(np);
+			} else if(event->atom == atoms[ATOM_NET_WM_STRUT]) {
+				ReadClientStrut(np);
 			}
 			break;
 		}
 
 		if(changed) {
 			DrawBorder(np);
-			DrawTray();
+			UpdateTaskBar();
+			UpdatePager();
 		}
-		if(np->statusFlags & STAT_WMDIALOG) {
+		if(np->state.status & STAT_WMDIALOG) {
 			return 0;
 		} else {
 			return 1;
@@ -486,20 +597,28 @@ int HandlePropertyNotify(const XPropertyEvent *event) {
 
 /****************************************************************************
  ****************************************************************************/
-void HandleConfigureNotify(const XConfigureEvent *event) {
+int HandleConfigureNotify(const XConfigureEvent *event) {
+
 #ifdef USE_SHAPE
+
 	ClientNode *np;
 
 	np = FindClientByWindow(event->window);
-	if(np && (np->statusFlags & STAT_USESHAPE)) {
+	if(np && (np->state.status & STAT_USESHAPE)) {
 		SetShape(np);
+		return 1;
 	}
+
 #endif
+
+	return 0;
+
 }
 
 /****************************************************************************
  ****************************************************************************/
 void HandleClientMessage(const XClientMessageEvent *event) {
+
 	ClientNode *np;
 	long mask, flags;
 	char *atomName;
@@ -521,11 +640,12 @@ void HandleClientMessage(const XClientMessageEvent *event) {
 
 			if(mask & WIN_STATE_HIDDEN) {
 				if(flags & WIN_STATE_HIDDEN) {
-					np->statusFlags |= STAT_NOLIST;
+					np->state.status |= STAT_NOLIST;
 				} else {
-					np->statusFlags &= ~STAT_NOLIST;
+					np->state.status &= ~STAT_NOLIST;
 				}
-				DrawTray();
+				UpdateTaskBar();
+				UpdatePager();
 			}
 
 		} else if(event->message_type == atoms[ATOM_WIN_LAYER]) {
@@ -554,13 +674,12 @@ void HandleClientMessage(const XClientMessageEvent *event) {
 
 		} else if(event->message_type == atoms[ATOM_NET_ACTIVE_WINDOW]) {
 
-			if(event->data.l[0] == 0) {
-				RestoreClient(np);
-			}
+			RestoreClient(np);
+			FocusClient(np);
 
 		} else if(event->message_type == atoms[ATOM_NET_WM_DESKTOP]) {
 
-			if(event->data.l[0] == 0xFFFFFFFF) {
+			if(event->data.l[0] == ~0L) {
 				SetClientSticky(np, 1);
 			} else {
 
@@ -568,17 +687,189 @@ void HandleClientMessage(const XClientMessageEvent *event) {
 					(np->controller)(0);
 				}
 
-				if(event->data.l[0] >= 0 && event->data.l[0] < desktopCount) {
-					np->statusFlags &= ~STAT_STICKY;
+				if(event->data.l[0] >= 0 && event->data.l[0] < (long)desktopCount) {
+					np->state.status &= ~STAT_STICKY;
 					SetClientDesktop(np, event->data.l[0]);
 				}
 			}
 
+		} else if(event->message_type == atoms[ATOM_NET_CLOSE_WINDOW]) {
+
+			DeleteClient(np);
+
+		} else if(event->message_type == atoms[ATOM_NET_MOVERESIZE_WINDOW]) {
+
+			HandleNetMoveResize(event, np);
+
+		} else if(event->message_type == atoms[ATOM_NET_WM_STATE]) {
+
+			HandleNetWMState(event, np);
+
 		} else {
+
 			atomName = JXGetAtomName(display, event->message_type);
 			Debug("Uknown ClientMessage: %s", atomName);
 			JXFree(atomName);
+
 		}
+
+	} else if(event->window == rootWindow) {
+
+		if(event->message_type == atoms[ATOM_JWM_RESTART]) {
+			Restart();
+		} else if(event->message_type == atoms[ATOM_JWM_EXIT]) {
+			Exit();
+		} else if(event->message_type == atoms[ATOM_NET_CURRENT_DESKTOP]) {
+			ChangeDesktop(event->data.l[0]);
+		}
+
+
+	} else if(event->message_type == atoms[ATOM_NET_SYSTEM_TRAY_OPCODE]) {
+
+		HandleDockEvent(event);
+
+	}
+
+}
+
+/****************************************************************************
+ * Handle a _NET_MOVERESIZE_WINDOW request.
+ ****************************************************************************/
+void HandleNetMoveResize(const XClientMessageEvent *event, ClientNode *np) {
+
+	long flags, gravity;
+	long x, y;
+	long width, height;
+	int deltax, deltay;
+	int north, south, east, west;
+
+	Assert(event);
+	Assert(np);
+
+	gravity = event->data.l[0] & 0xFF;
+	flags = event->data.l[0] >> 8;
+
+	x = np->x;
+	y = np->y;
+	width = np->width;
+	height = np->height;
+
+	if(flags & (1 << 0)) {
+		x = event->data.l[1];
+	}
+	if(flags & (1 << 1)) {
+		y = event->data.l[2];
+	}
+	if(flags & (1 << 2)) {
+		width = event->data.l[3];
+	}
+	if(flags & (1 << 3)) {
+		height = event->data.l[4];
+	}
+
+	if(gravity == 0) {
+		gravity = np->gravity;
+	}
+
+	GetBorderSize(np, &north, &south, &east, &west);
+	GetGravityDelta(gravity, north, south, east, west, &deltax, &deltay);
+
+	x -= deltax;
+	y -= deltay;
+
+	np->x = x;
+	np->y = y;
+	np->width = width;
+	np->height = height;
+
+	JXMoveResizeWindow(display, np->parent,
+		np->x - west, np->y - north,
+		np->width + east + west,
+		np->height + north + south);
+	JXMoveResizeWindow(display, np->window, west, north,
+		np->width, np->height);
+
+	WriteState(np);
+	SendConfigureEvent(np);
+
+}
+
+/****************************************************************************
+ * Handle a _NET_WM_STATE request.
+ ****************************************************************************/
+void HandleNetWMState(const XClientMessageEvent *event, ClientNode *np) {
+
+	int actionMaximize;
+	int actionStick;
+	int actionShade;
+	int x;
+
+	/* Up to two actions to be applied together, figure it out. */
+	actionMaximize = 0;
+	actionStick = 0;
+	actionShade = 0;
+
+	for(x = 1; x <= 2; x++) {
+		if(event->data.l[x]
+			== (long)atoms[ATOM_NET_WM_STATE_STICKY]) {
+			actionStick = 1;
+		} else if(event->data.l[x]
+			== (long)atoms[ATOM_NET_WM_STATE_MAXIMIZED_VERT]) {
+			actionMaximize = 1;
+		} else if(event->data.l[x]
+			== (long)atoms[ATOM_NET_WM_STATE_MAXIMIZED_HORZ]) {
+			actionMaximize = 1;
+		} else if(event->data.l[x]
+			== (long)atoms[ATOM_NET_WM_STATE_SHADED]) {
+			actionShade = 1;
+		}
+	}
+
+	switch(event->data.l[0]) {
+	case 0: /* Remove */
+		if(actionStick) {
+			SetClientSticky(np, 0);
+		}
+		if(actionMaximize && (np->state.status & STAT_MAXIMIZED)) {
+			MaximizeClient(np);
+		}
+		if(actionShade) {
+			UnshadeClient(np);
+		}
+		break;
+	case 1: /* Add */
+		if(actionStick) {
+			SetClientSticky(np, 1);
+		}
+		if(actionMaximize && !(np->state.status & STAT_MAXIMIZED)) {
+			MaximizeClient(np);
+		}
+		if(actionShade) {
+			ShadeClient(np);
+		}
+		break;
+	case 2: /* Toggle */
+		if(actionStick) {
+			if(np->state.status & STAT_STICKY) {
+				SetClientSticky(np, 0);
+			} else {
+				SetClientSticky(np, 1);
+			}
+		}
+		if(actionMaximize) {
+			MaximizeClient(np);
+		}
+		if(actionShade) {
+			if(np->state.status & STAT_SHADED) {
+				UnshadeClient(np);
+			} else {
+				ShadeClient(np);
+			}
+		}
+		break;
+	default:
+		Debug("bad _NET_WM_STATE action: %ld", event->data.l[0]);
+		break;
 	}
 }
 
@@ -594,7 +885,7 @@ void HandleMotionNotify(const XMotionEvent *event) {
 	}
 
 	np = FindClientByParent(event->window);
-	if(np && (np->borderFlags & BORDER_OUTLINE)) {
+	if(np && (np->state.border & BORDER_OUTLINE)) {
 		action = GetBorderActionType(np, event->x, event->y);
 		if(np->borderAction != action) {
 			np->borderAction = action;
@@ -609,12 +900,14 @@ void HandleMotionNotify(const XMotionEvent *event) {
  ****************************************************************************/
 #ifdef USE_SHAPE
 void HandleShapeEvent(const XShapeEvent *event) {
+
 	ClientNode *np;
 
 	np = FindClientByWindow(event->window);
 	if(np) {
 		SetShape(np);
 	}
+
 }
 #endif /* USE_SHAPE */
 
@@ -630,6 +923,7 @@ void HandleColormapChange(const XColormapEvent *event) {
 			UpdateClientColormap(np);
 		}
 	}
+
 }
 
 /****************************************************************************
@@ -651,20 +945,21 @@ void HandleMapRequest(const XMapEvent *event) {
 		} else {
 			JXMapWindow(display, event->window);
 		}
-		JXUngrabServer(display);
 		JXSync(display, False);
+		JXUngrabServer(display);
 	} else {
-		if(!(np->statusFlags & STAT_MAPPED)) {
-			np->statusFlags |= STAT_MAPPED;
-			np->statusFlags &= ~STAT_MINIMIZED;
-			np->statusFlags &= ~STAT_WITHDRAWN;
+		if(!(np->state.status & STAT_MAPPED)) {
+			np->state.status |= STAT_MAPPED;
+			np->state.status &= ~STAT_MINIMIZED;
+			np->state.status &= ~STAT_WITHDRAWN;
 			JXMapWindow(display, np->window);
 			JXMapWindow(display, np->parent);
 			RaiseClient(np);
 			if(focusModel == FOCUS_CLICK) {
 				FocusClient(np);
 			}
-			DrawTray();
+			UpdateTaskBar();
+			UpdatePager();
 		}
 	}
 	RestackClients();
@@ -684,18 +979,23 @@ void HandleUnmapNotify(const XUnmapEvent *event) {
 			(np->controller)(1);
 		}
 
-		if(np->statusFlags & STAT_MAPPED) {
-			np->statusFlags &= ~STAT_MAPPED;
+		if(np->state.status & STAT_MAPPED) {
+			np->state.status &= ~STAT_MAPPED;
 			JXUnmapWindow(display, np->parent);
 		}
 
+	} else if(!np) {
+
+		HandleDockDestroy(event->window);
+
 	}
 
 }
 
 /****************************************************************************
  ****************************************************************************/
-void HandleDestroyNotify(const XDestroyWindowEvent *event) {
+int HandleDestroyNotify(const XDestroyWindowEvent *event) {
+
 	ClientNode *np;
 
 	np = FindClientByWindow(event->window);
@@ -706,22 +1006,28 @@ void HandleDestroyNotify(const XDestroyWindowEvent *event) {
 		}
 
 		RemoveClient(np);
+
+		return 1;
+
+	} else if(!np) {
+
+		return HandleDockDestroy(event->window);
+
 	}
 
+	return 0;
+
 }
 
 /****************************************************************************
  ****************************************************************************/
 void DispatchBorderButtonEvent(const XButtonEvent *event, ClientNode *np) {
+
 	static Time lastClickTime = 0;
 	static int lastX = 0, lastY = 0;
 	static int doubleClickActive = 0;
 	BorderActionType action;
 
-	if(!(np->borderFlags & BORDER_OUTLINE)) {
-		return;
-	}
-
 	action = GetBorderActionType(np, event->x, event->y);
 
 	switch(action & 0x0F) {
@@ -736,7 +1042,7 @@ void DispatchBorderButtonEvent(const XButtonEvent *event, ClientNode *np) {
 				&& abs(event->time - lastClickTime) <= doubleClickSpeed
 				&& abs(event->x - lastX) <= doubleClickDelta
 				&& abs(event->y - lastY) <= doubleClickDelta) {
-				if(np->statusFlags & STAT_SHADED) {
+				if(np->state.status & STAT_SHADED) {
 					UnshadeClient(np);
 				} else {
 					ShadeClient(np);
@@ -775,3 +1081,19 @@ void DispatchBorderButtonEvent(const XButtonEvent *event, ClientNode *np) {
 
 }
 
+/****************************************************************************
+ ****************************************************************************/
+void SetShowMenuOnRoot(const char *mask) {
+
+	int x;
+
+	Assert(mask);
+
+	onRootMask = 0;
+	for(x = 0; mask[x]; x++) {
+		onRootMask |= 1 << (mask[x] - '0');
+	}
+
+}
+
+
