@@ -6,8 +6,12 @@
 #ifndef EVENT_H
 #define EVENT_H
 
+void HandleStartupEvents();
+
 void WaitForEvent();
 void ProcessEvent(XEvent *event);
 
+void SetShowMenuOnRoot(const char *mask);
+
 #endif
 
