@@ -13,6 +13,8 @@
 #define Round( x ) ( (long)((x) + 0.5) )
 
 void ExpandPath(char **path);
+void Trim(char *str);
+char *CopyString(const char *str);
 
 #endif
 
