@@ -6,14 +6,16 @@
 #ifndef PAGER_H
 #define PAGER_H
 
-extern int pagerWidth;
+struct TrayComponentType;
 
 void InitializePager();
 void StartupPager();
 void ShutdownPager();
 void DestroyPager();
 
-void DrawPager(Window w, GC gc, int xoffset);
+struct TrayComponentType *CreatePager();
+
+void UpdatePager();
 
 #endif
 
