@@ -8,6 +8,13 @@
 
 #define ZERO_TIME { 0, 0 }
 
+typedef struct TimeType {
+
+	unsigned long seconds;
+	int ms;
+
+} TimeType;
+
 void InitializeTiming();
 void StartupTiming();
 void ShutdownTiming();
@@ -17,10 +24,7 @@ void GetCurrentTime(TimeType *t);
 
 unsigned long GetTimeDifference(const TimeType *t1, const TimeType *t2);
 
-char *GetShortTimeString();
-char *GetLongTimeString();
-
-void SetClockFormat(const char *f);
+const char *GetTimeString(const char *format);
 
 #endif
 
