@@ -6,7 +6,10 @@
 #ifndef WINMENU_H
 #define WINMENU_H
 
-void ShowWindowMenu(ClientNode *np, int x, int y);
+struct ClientNode;
+
+void GetWindowMenuSize(struct ClientNode *np, int *width, int *height);
+void ShowWindowMenu(struct ClientNode *np, int x, int y);
 
 #endif
 
